@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1998--2005  Guido Masarotto and Brian Ripley
- *  Copyright (C) 2004--2018  The R Foundation
+ *  Copyright (C) 2004--2021  The R Foundation
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -673,7 +673,7 @@ static void menuabout(control m)
 
 static void menuRhome(control m)
 {
-    ShellExecute(NULL, "open", "http://www.r-project.org", NULL, NULL, SW_SHOW);
+    ShellExecute(NULL, "open", "https://www.r-project.org", NULL, NULL, SW_SHOW);
 }
 
 static void menuCRAN(control m)
@@ -1026,8 +1026,10 @@ int setupui(void)
     /* set locale before doing anything with menus */
     setlocale(LC_CTYPE, ""); /* necessary in case next fails to set
 				a valid locale */
-    if((p = getenv("LC_ALL"))) strncpy(Rlocale, p, sizeof(Rlocale)-1);
-    if((p = getenv("LC_CTYPE"))) strncpy(Rlocale, p, sizeof(Rlocale)-1);
+    if((p = getenv("LC_ALL")))
+	strncpy(Rlocale, p, sizeof(Rlocale)-1);
+    else if((p = getenv("LC_CTYPE")))
+	strncpy(Rlocale, p, sizeof(Rlocale)-1);
     if (strcmp(Rlocale, "C") == 0) strcpy(Rlocale, "en");
     setlocale(LC_CTYPE, Rlocale);
     mbcslocale = MB_CUR_MAX > 1;
@@ -1363,7 +1365,7 @@ int winaddmenu(const char *name, char *errmsg)
     }
     if (m) {
 	usermenus[nmenus] = m;
-	usermenunames[nmenus] = strdup(name);
+	usermenunames[nmenus] = strdup(name); // FIXME: allocation could fail.
 	nmenus++;
 	show(RConsole);
 	return 0;
