@@ -1,7 +1,7 @@
 #  File src/library/tools/R/news.R
-#  Part of the R package, http://www.R-project.org
+#  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2013 The R Core Team
+#  Copyright (C) 1995-2020 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
 #  GNU General Public License for more details.
 #
 #  A copy of the GNU General Public License is available at
-#  http://www.r-project.org/Licenses/
+#  https://www.R-project.org/Licenses/
 
 
 ## .build_news_db_from_R_NEWS <-
@@ -24,7 +24,7 @@
 ##     ## This currently is a list of x.y lists of x.y.z lists of
 ##     ## categories list of entries.
 ##     flatten <- function(e)
-##         cbind(rep.int(names(e), sapply(e, length)),
+##         cbind(rep.int(names(e), lengths(e)),
 ##               unlist(lapply(e,
 ##                             function(s) {
 ##                                 ## Also remove leading white space and
@@ -59,14 +59,19 @@ function(package, lib.loc = NULL, format = NULL, reader = NULL)
     ## little point in providing format/reader support ...
     ## </FIXME>
 
-    ## Look for new-style inst/NEWS.Rd installed as NEWS.Rd.
-    ## If not found, look for old-style
+    ## Look for new-style inst/NEWS.Rd installed as NEWS.Rd
+    ## If not found, look for NEWS.md.
+    ## If not found, look at old-style
     ##   NEWS inst/NEWS
     ## installed as NEWS (and ignore ChangeLog files).
     nfile <- file.path(dir, "NEWS.Rd")
     if(file_test("-f", nfile))
         return(.build_news_db_from_package_NEWS_Rd(nfile))
 
+    nfile <- file.path(dir, "NEWS.md")
+    if(file_test("-f", nfile))
+        return(.build_news_db_from_package_NEWS_md(nfile))
+
     nfile <- file.path(dir, "NEWS")
     if(!file_test("-f", nfile))
         return(invisible())
@@ -106,7 +111,7 @@ function(file)
         if(!is.na(encoding))
             lines[ind] <- iconv(lines[ind], encoding, "")
         ## Last resort.
-        if(any(is.na(nchar(lines[ind], allowNA = TRUE))))
+        if(anyNA(nchar(lines[ind], allowNA = TRUE)))
             lines[ind] <- iconv(lines[ind], "", "", sub = "byte")
     }
 
@@ -121,9 +126,9 @@ function(file)
     ## liberal.
     re_valid_package_name <- .standard_regexps()$valid_package_name
     re_v <- sprintf("^([[:space:]]*(%s)|(%s))(%s).*$",
-                    paste("CHANGES? *(IN|FOR).*VERSION *",
-                          "CHANGES? *(IN|FOR|TO) *",
-                          sep = "|"),
+                    paste0("CHANGES? *(IN|FOR).*VERSION *",
+                           "|",
+                           "CHANGES? *(IN|FOR|TO) *"),
                     sprintf(paste(## TeachingDemos pomp ouch
                                   "NEW IN .*",
                                   ## HyperbolicDist nls2 proto
@@ -286,9 +291,7 @@ function(file)
     exdent <- exdent[exdent > 1L]
     if(length(exdent)) {
         out[, 4L] <-
-            gsub(sprintf("\n%s",
-                         paste(rep.int(" ", min(exdent) - 1L),
-                               collapse = "")),
+            gsub(sprintf("\n%s", strrep(" ", min(exdent) - 1L)),
                  "\n", entries)
     }
 
@@ -298,16 +301,17 @@ function(file)
 .make_news_db <-
 function(x, bad = NULL, classes = NULL)
 {
-    ## Expect x to be a 4 column
+    ## Expect x to be a character matrix giving at least
     ##   version date category text
-    ## character matrix.
+    ## in its first 4 columns.
     ## Could of course check for this using
-    ##   if(!is.character(x) || ncol(x) != 4L)
+    ##   if(!is.character(x) || ncol(x) < 4L)
     out <- data.frame(x, row.names = NULL, stringsAsFactors = FALSE)
     ## Note that we cannot do
     ##   dimnames(out) <- list(NULL,
     ##                         c("Version", "Date", "Category", "Text"))
-    colnames(out) <- c("Version", "Date", "Category", "Text")
+    colnames(out)[1L : 4L] <-
+        c("Version", "Date", "Category", "Text")
     if(!is.null(bad))
         attr(out, "bad") <- bad
     class(out) <- unique(c(classes, "news_db", "data.frame"))
@@ -323,23 +327,25 @@ Rd2txt_NEWS_in_Rd_options <-
 
 Rd2txt_NEWS_in_Rd <-
 function(f, out = "") {
-    if (grepl("[.]rds$", f)) f <- readRDS(f)
+    if (endsWith(f, ".rds")) f <- readRDS(f)
     Rd2txt(f, out,
            stages = c("install", "render"),
            outputEncoding = if(l10n_info()[["UTF-8"]]) "" else "ASCII//TRANSLIT",
-           options = Rd2txt_NEWS_in_Rd_options)
+           options = Rd2txt_NEWS_in_Rd_options,
+           macros = file.path(R.home("share"), "Rd", "macros", "system.Rd"))
  }
 
 Rd2HTML_NEWS_in_Rd <-
 function(f, out, ...) {
-    if (grepl("[.]rds$", f)) f <- readRDS(f)
-    Rd2HTML(f, out, stages = c("install", "render"), ...)
+    if (endsWith(f, ".rds")) f <- readRDS(f)
+    Rd2HTML(f, out, stages = c("install", "render"),
+           macros = file.path(R.home("share"), "Rd", "macros", "system.Rd"), ...)
 }
 
 Rd2pdf_NEWS_in_Rd <-
 function(f, pdf_file)
 {
-    if (grepl("[.]rds$", f)) f <- readRDS(f)
+    if (endsWith(f, ".rds")) f <- readRDS(f)
     f2 <- tempfile()
     ## See the comments in ?texi2dvi about spaces in paths
     f3 <- if(grepl(" ", Sys.getenv("TMPDIR")))
@@ -349,14 +355,15 @@ function(f, pdf_file)
     out <- file(f3, "w")
     Rd2latex(f, f2,
              stages = c("install", "render"),
-             outputEncoding = "UTF-8", writeEncoding = FALSE)
+             outputEncoding = "UTF-8", writeEncoding = FALSE,
+             macros = file.path(R.home("share"), "Rd", "macros", "system.Rd"))
     cat("\\documentclass[", Sys.getenv("R_PAPERSIZE"), "paper]{book}\n",
         "\\usepackage[ae,hyper]{Rd}\n",
         "\\usepackage[utf8]{inputenc}\n",
         "\\usepackage{graphicx}\n",
         "\\setkeys{Gin}{width=0.7\\textwidth}\n",
-        "\\graphicspath{{", normalizePath(file.path(R.home("doc"), "html"), "/"),
-                            "/}}\n",
+        "\\graphicspath{{\"", normalizePath(file.path(R.home("doc"), "html"), "/"),
+                            "/\"}}\n",
         "\\hypersetup{pdfpagemode=None,pdfstartview=FitH}\n",
         "\\begin{document}\n",
         "\\chapter*{}\\sloppy\n",
@@ -369,7 +376,9 @@ function(f, pdf_file)
     close(out)
     od <- setwd(dirname(f3))
     on.exit(setwd(od))
-    texi2pdf("NEWS.tex", quiet = TRUE)
+    ## avoid broken texi2pdf scripts: this is simple LaTeX
+    ## and emulation suffices
+    texi2pdf("NEWS.tex", quiet = TRUE, texi2dvi = "emulation")
     setwd(od); on.exit()
     invisible(file.copy(file.path(dirname(f3), "NEWS.pdf"),
                         pdf_file, overwrite = TRUE))
@@ -386,16 +395,34 @@ function(file, out = stdout(), codify = FALSE)
     ## encoding.
 
     file <- file_path_as_absolute(file)
-    dir <- dirname(file)
-    format <- "default"
-    if(file_test("-f", dfile <- file.path(dir, "DESCRIPTION")))
-        meta <- .read_description(dfile)
-    else if(basename(dir) == "inst" &&
-            file_test("-f", dfile <- file.path(dirname(dir),
-                                               "DESCRIPTION")))
-        meta <- .read_description(dfile)
-    else
-        format <- "R"
+
+    if(file_test("-d", file)) {
+        dir <- file
+        dfile <- file.path(dir, "DESCRIPTION")
+        if(!file_test("-f", dfile))
+            stop("DESCRIPTION file not found")
+        file <- file.path(dir, "inst", "NEWS")
+        if(!file_test("-f", file)) {
+            file <- file.path(dir, "NEWS")
+            if(!file_test("-f", file))
+                stop("NEWS file not found")
+        }
+    } else {
+        dir <- dirname(file)
+        dfile <- file.path(dir, "DESCRIPTION")
+        if(!file_test("-f", dfile)) {
+            if((basename(dir) != "inst") ||
+               !file_test("-f",
+                          dfile <- file.path(dirname(dir),
+                                             "DESCRIPTION")))
+                stop("DESCRIPTION file not found")
+        }
+    }
+
+    ## No longer support taking NEWS files without correponding
+    ## DESCRIPTION file as being from R itself (PR #16556).
+
+    meta <- .read_description(dfile)
 
     wto <- function(x) writeLines(x, con = out, useBytes = TRUE)
     cre <- "(\\W|^)(\"[[:alnum:]_.]*\"|[[:alnum:]_.:]+\\(\\))(\\W|$)"
@@ -409,38 +436,9 @@ function(file, out = stdout(), codify = FALSE)
         on.exit(close(out))
     }
 
-    if(format == "R") {
-        news <- readNEWS(chop = "keepAll")
-        if(!length(news))
-            stop("No news found in given file using old-style R-like format.")
-        wto(c("\\newcommand{\\PR}{\\Sexpr[results=rd]{tools:::Rd_expr_PR(#1)}}",
-              "\\name{NEWS}",
-              "\\title{R News}",
-              "\\encoding{UTF-8}"))
-        for(y in news) {
-            for(i in seq_along(y)) {
-                wto(sprintf("\\section{CHANGES IN R VERSION %s}{",
-                            names(y)[i]))
-                z <- y[[i]]
-                for(j in seq_along(z)) {
-                    wto(c(sprintf("  \\subsection{%s}{", names(z)[j]),
-                          "    \\itemize{"))
-                    for(chunk in z[[j]]) {
-                        chunk <- toRd(paste(chunk, collapse = "\n      "))
-                        if(codify) {
-                            chunk <- gsub(cre, "\\1\\\\code{\\2}\\3",
-                                          chunk)
-                        }
-                        chunk <- gsub("PR#([[:digit:]]+)", "\\\\PR{\\1}",
-                                      chunk)
-                        wto(paste("      \\item", enc2utf8(chunk)))
-                    }
-                    wto(c("    }", "  }"))
-                }
-                wto("}")
-            }
-        }
-    } else {
+    ## had   if(format == "R") {
+    ## and this was   } else { format == "default" :
+    {
         news <- .news_reader_default(file)
         bad <- attr(news, "bad")
         if(!length(bad))
@@ -496,14 +494,14 @@ function(file, out = stdout(), codify = FALSE)
                           "    \\itemize{",
                           paste("      \\item",
                                 gsub("\n", "\n        ",
-                                     cchunks[[j]]$Text)),
+                                     cchunks[[j]]$Text, fixed=TRUE)),
                           "    }",
                           "  }"))
                 }
             } else {
                 wto(c("  \\itemize{",
                       paste("    \\item",
-                            gsub("\n", "\n      ", vchunk$Text)),
+                            gsub("\n", "\n      ", vchunk$Text, fixed=TRUE)),
                       "  }"))
             }
             wto("}")
@@ -511,73 +509,75 @@ function(file, out = stdout(), codify = FALSE)
     }
 }
 
-Rd_expr_PR <-
-function(x)
-{
-    baseurl <- "https://bugs.R-project.org/bugzilla3/show_bug.cgi?id"
-    sprintf("\\href{%s=%s}{PR#%s}", baseurl, x, x)
-}
-
 .build_news_db_from_R_NEWS_Rd <-
-function(file = NULL)
+function(file = NULL, Rfile = "NEWS.rds")
 {
     x <- if(is.null(file))
-        readRDS(file.path(R.home("doc"), "NEWS.rds"))
+        readRDS(file.path(R.home("doc"), Rfile))
     else {
         ## Expand \Sexpr et al now because this does not happen when using
         ## fragments.
-        prepare_Rd(parse_Rd(file), stages = "install")
+        macros <- initialRdMacros()
+        prepare_Rd(parse_Rd(file, macros = macros), stages = "install")
     }
 
     db <- .extract_news_from_Rd(x)
-    db <- db[db[,1L] != "CHANGES in previous versions",,drop = FALSE]
+    skip <- c("CHANGES in previous versions", "LATER NEWS", "OLDER NEWS")
+    db <- db[!(db[,1L] %in% skip),,drop = FALSE]
 
     ## Squeeze in an empty date column.
     .make_news_db(cbind(sub("^CHANGES IN (R )?(VERSION )?", "", db[, 1L]),
                         NA_character_,
                         db[, 2L],
-                        sub("\n*$", "", db[, 3L])),
-                  logical(nrow(db)),
+                        Text = sub("\n*$", "", db[, 3L]),
+                        HTML = db[, 4L]),
+                  NULL,
                   "news_db_from_Rd")
 }
 
 .build_news_db_from_package_NEWS_Rd <-
 function(file)
 {
-    x <- prepare_Rd(parse_Rd(file), stages = "install")
+    macros <- initialRdMacros()
+    x <- prepare_Rd(parse_Rd(file, macros = macros), stages = "install")
 
     db <- .extract_news_from_Rd(x)
 
     ## Post-process section names to extract versions and dates.
     re_v <- sprintf(".*version[[:space:]]+(%s).*$",
                     .standard_regexps()$valid_package_version)
-    re_d <- sprintf("^.*(%s)[[:punct:][:space:]]*$",
-                    "[[:digit:]]{4}-[[:digit:]]{2}-[[:digit:]]{2}")
-
+    reDt <- "[[:digit:]]{4}-[[:digit:]]{2}-[[:digit:]]{2}"
+    rEnd <- "[[:punct:][:space:]]*$"
+    re_d1 <- sprintf(paste0("^.*(%s)", rEnd), reDt)
+    ## or ending with '(YYYY-MM-DD, <note>)'
+    re_d2 <- sprintf(paste0("^.*\\((%s)[[:punct:]] .*\\)", rEnd), reDt)
     nms <- db[, 1L]
     ind <- grepl(re_v, nms, ignore.case = TRUE)
     if(!all(ind))
         warning("Cannot extract version info from the following section titles:\n",
-                sprintf("  %s", unique(nms[!ind])))
+		paste(unique(nms[!ind]), collapse = "  "))
     .make_news_db(cbind(ifelse(ind,
-                               sub(re_v, "\\1", nms, ignore.case = TRUE),
-                               NA_character_),
-                        ifelse(grepl(re_d, nms, perl = TRUE),
-                               sub(re_d, "\\1", nms, perl = TRUE),
-                               NA_character_),
-                        db[, 2L],
-                        sub("\n*$", "", db[, 3L])),
-                  logical(nrow(db)),
+			       sub(re_v, "\\1", nms, ignore.case = TRUE),
+			       NA_character_),
+			ifelse(grepl(re_d1, nms, perl = TRUE),
+			       sub(re_d1, "\\1", nms, perl = TRUE),
+			       ifelse(grepl(re_d2, nms, perl = TRUE),
+				      sub(re_d2, "\\1", nms, perl = TRUE),
+				      NA_character_)),
+			db[, 2L],
+                        Text = sub("\n*$", "", db[, 3L]),
+                        HTML = db[, 4L]),
+                  NULL,
                   "news_db_from_Rd")
 }
 
 .extract_news_from_Rd <-
 function(x)
 {
-    .get_Rd_section_names <- function(x)
+    get_section_names <- function(x)
         sapply(x, function(e) .Rd_get_text(e[[1L]]))
 
-    do_chunk <- function(x) {
+    get_item_texts <- function(x) {
         ## Currently, chunks should consist of a single \itemize list
         ## containing the news items.  Notify if there is more than one
         ## such list, and stop if there is none.
@@ -585,67 +585,281 @@ function(x)
         pos <- which(RdTags(x) == "\\itemize")
         if(!length(pos)) {
             stop(gettextf("Malformed NEWS.Rd file:\nChunk starting\n  %s\ncontains no \\itemize.",
-                          substring(sub("^[[:space:]]*", "",
-                                        .Rd_deparse(x)),
-                                    1L, 60L)),
+                          substr(sub("^[[:space:]]*", "",
+                                     .Rd_deparse(x)),
+                                 1L, 60L)),
                  domain = NA)
         } else if(length(pos) > 1L) {
             warning(gettextf("Malformed NEWS.Rd file:\nChunk starting\n  %s\ncontains more than one \\itemize.\nUsing the first one.",
-                             substring(sub("^[[:space:]]*", "",
-                                           .Rd_deparse(x)),
-                                       1L, 60L)),
+                             substr(sub("^[[:space:]]*", "",
+                                        .Rd_deparse(x)),
+                                    1L, 60L)),
                     domain = NA)
             pos <- pos[1L]
         }
-        x <- x[pos]
 
-        out <- NULL
-        zz <- textConnection("out", "w", local = TRUE)
-        on.exit(close(zz))
-        Rd2txt(x, out = zz, fragment = TRUE,
-               options =
-               c(Rd2txt_NEWS_in_Rd_options,
-                 list(itemBullet = "\036  ")))
+        x <- x[[pos]]
 
-        ## Try to find the column offset of the top-level bullets.
-        pat <- "^( *)\036.*"
-        off <- min(nchar(sub(pat, "\\1", out[grepl(pat, out)])))
-        pat <- sprintf("^%s\036 ",
-                       paste(rep.int(" ", off), collapse = ""))
-        s <- sub(pat, "\036", out)
-        ## Try to remove some indent for nested material.
-        pat <- sprintf("^%s",
-                       paste(rep.int(" ", off + 2L), collapse = ""))
-        s <- sub(pat, "", s)
+        out <- file()
+        on.exit(close(out))
 
-        s <- paste(s, collapse = "\n")
-        s <- sub("^[[:space:]]*\036", "", s)
-        s <- sub("[[:space:]]*$", "", s)
-        ## <FIXME>
-        ## Could be more fancy and use \u2022 "if possible".
-        gsub("\036", "*", unlist(strsplit(s, "\n\036", fixed = TRUE)))
-        ## </FIXME>
-    }
+        Rd2txt_options <- Rd2txt_NEWS_in_Rd_options
+        Rd2txt_options$width <- 72L
 
-    y <- x[RdTags(x) == "\\section"]
-    do.call(rbind,
-            Map(cbind,
-                .get_Rd_section_names(y),
+        ## Extract and process \item chunks:
+        y <- split(x, cumsum(RdTags(x) == "\\item"))
+        y <- y[names(y) != "0"]
+        if(!length(y)) {
+            warning(gettextf("Malformed NEWS.Rd file:\nChunk starting\n  %s\ncontains no \\item.",
+                             substr(sub("^[[:space:]]*", "",
+                                        .Rd_deparse(x)),
+                                    1L, 60L)),
+                    domain = NA)
+            return(matrix(character(), 0L, 2L,
+                          dimnames = list(NULL, c("Text", "HTML"))))
+        }
+        do.call(rbind,
                 lapply(y,
                        function(e) {
-                           z <- e[[2L]]
-                           ind <- RdTags(z) == "\\subsection"
-                           if(any(ind)) {
-                               z <- z[ind]
-                               do.call(rbind,
-                                       Map(cbind,
-                                           .get_Rd_section_names(z),
-                                           lapply(z,
-                                                  function(e)
-                                                  do_chunk(e[[2L]]))))
-                           } else {
-                               cbind(NA_character_, do_chunk(z))
-                           }
-                       })))
+                           ## Drop \item.
+                           e <- e[-1L]
+                           ## Convert to text.
+                           Rd2txt(e, fragment = TRUE, out = out,
+                                  options = Rd2txt_options)
+                           one <- paste(readLines(out, warn = FALSE),
+                                        collapse = "\n")
+                           ## Need warn = FALSE to avoid warning about
+                           ## incomplete final line for e.g. 'cluster'.
+                           ## Convert to HTML.
+                           Rd2HTML(e, fragment = TRUE, out = out)
+                           two <- paste(readLines(out, warn = FALSE),
+                                        collapse = "\n")
+                           cbind(Text = one, HTML = two)
+                       }))
+    }
+
+    cbind_safely <- function(u, v)
+        cbind(rep_len(u, NROW(v)), v)
+
+    x <- x[RdTags(x) == "\\section"]
+    y <- Map(cbind_safely,
+             get_section_names(x),
+             lapply(x,
+                    function(e) {
+                        z <- e[[2L]]
+                        ind <- RdTags(z) == "\\subsection"
+                        if(any(ind)) {
+                            z <- z[ind]
+                            do.call(rbind,
+                                    Map(cbind_safely,
+                                        get_section_names(z),
+                                        lapply(z,
+                                               function(e)
+                                                   get_item_texts(e[[2L]]))))
+                        } else {
+                            cbind_safely(NA_character_,
+                                         get_item_texts(z))
+                        }
+                    }))
+    y <- do.call(rbind, y)
+    ## Sanitze HTML.
+    s <- trimws(y[, "HTML"])
+    i <- startsWith(s, "<p>") & !endsWith(s, "</p>")
+    s[i] <- paste0(s[i], "</p>")
+    y[, "HTML"] <- s
+
+    y
 
 }
+
+.build_news_db_from_package_NEWS_md <-
+function(f)
+{
+    md <- readLines(f, encoding = "UTF-8", warn = FALSE)
+
+    ## Handle YAML header.
+    if(md[1L] == "---") {
+        for(pos in seq.int(2L, length(md)))
+            if(md[pos] == "---") break
+        md[seq_len(pos)] <- ""
+    }
+
+    doc <- commonmark::markdown_xml(md,
+                                    extensions = TRUE,
+                                    sourcepos = TRUE)
+    doc <- xml2::xml_ns_strip(xml2::read_xml(doc))
+
+    nodes <- xml2::xml_children(doc)    # Need xml2::xml_root()?
+
+    ## Inline for efficiency.
+    .markdown_text <- commonmark::markdown_text
+    .markdown_html <- commonmark::markdown_html
+    .xml_attr <- xml2::xml_attr
+    .xml_name <- xml2::xml_name
+    .xml_text <- xml2::xml_text
+
+    get_text_and_HTML <- function(sp) {
+        ## Sourcepos sp already split into l1 c2 l2 c2, for legibility:
+        l1 <- sp[1L]; c1 <- sp[2L]; l2 <- sp[3L]; c2 <- sp[4L]
+        txt <- if(l1 < l2) {
+                   c(substring(md[l1], c1),
+                     md[seq.int(from = l1 + 1L,
+                                length.out = l2 - l1 - 1L)],
+                     substring(md[l2], 1L, c2))
+               } else
+                   substring(md[l1], c1, c2)
+        c(.markdown_text(txt, width = 72L),
+          .markdown_html(txt))
+    }
+
+    do_vchunk <- function(nodes) {
+        ## Get version and date from heading.
+        version <- .xml_text(nodes[[1L]])
+        nodes <- nodes[-1L]
+        if(!length(nodes))
+            return(rbind(c(version, "", "", "")))
+        ## Unlike news in Rd where we (currently) insist on all news to
+        ## be given as items in itemize lists, for md we only split news
+        ## in version chunks according to category.  If the chunks has
+        ## headings, we take those with the same level as the first one
+        ## to start category chunks, and everything before the first
+        ## such heading as a chunk with an empty category (empty instead
+        ## of missing to make querying more convenient).  If there are
+        ## no headings, we have a single version chunk with no (empty)
+        ## category.
+        ind <- .xml_name(nodes) == "heading"
+        pos <- which(ind)
+        if(length(pos)) {
+            lev <- .xml_attr(nodes[pos], "level")
+            ind[pos] <- (lev == lev[1L])
+            if((pos[1L]) > 1L) {
+                ini <- seq_len(pos[1L] - 1L)
+                out <- list(do_cchunk(nodes[ini], FALSE))
+                nodes <- nodes[-ini]
+                ind <- ind[-ini]
+            } else
+                out <- list()
+            out <- c(out,
+                     lapply(split(nodes, cumsum(ind)),
+                            do_cchunk, TRUE))
+            cbind(version, do.call(rbind, out))
+        } else {
+            rbind(c(version,
+                    do_cchunk(nodes, FALSE)))
+        }
+
+    }
+
+    do_cchunk <- function(nodes, heading) {
+        ## See above: if the category chunk has a heading, we extract
+        ## the category from it.  Otherwise, the category is empty.
+        if(heading) {
+            category <- .xml_text(nodes[[1L]])
+            nodes <- nodes[-1L]
+        } else {
+            category <- ""
+        }
+
+        if(!length(nodes))
+            return(c(category, "", ""))
+
+        ## Compute text and HTML by converting everything from the start
+        ## of the first sourcepos to the end of the last sourcepos.
+        sp <- c(.xml_attr(nodes[[1L]], "sourcepos"),
+                .xml_attr(nodes[[length(nodes)]], "sourcepos"))
+        ## (If there is one node, nodes[c(1L, length(nodes))] would give
+        ## that node only once.  Could also special case ...)
+        sp <- as.integer(unlist(strsplit(sp, "[:-]"))[c(1L, 2L, 7L, 8L)])
+
+        c(category, get_text_and_HTML(sp))
+    }
+
+    ind <- .xml_name(nodes) == "heading"
+    pos <- which(ind)
+    if(!length(pos)) return()
+
+    ## Skip leading headings until we find one from which we can extract
+    ## a version number.  Then drop everything ahead of this, and take
+    ## all headings with the same level to start version chunks.
+
+    re_v <- sprintf("(^|.*[[:space:]]+)[vV]?(%s).*$",
+                    .standard_regexps()$valid_package_version)
+    while(length(pos) &&
+          !grepl(re_v, .xml_text(nodes[[pos[1L]]])))
+        pos <- pos[-1L]
+    if(!length(pos)) return()
+
+    lev <- .xml_attr(nodes[pos], "level")
+    ind[pos] <- (lev == lev[1L])
+    if(pos[1L] > 1L) {
+        ini <- seq_len(pos[1L] - 1L)
+        nodes <- nodes[-ini]
+        ind <- ind[-ini]
+    }
+    vchunks <- split(nodes, cumsum(ind))
+    db <- do.call(rbind, lapply(vchunks, do_vchunk))
+
+    ## Very similar to .build_news_db_from_package_NEWS_Rd() ...
+
+    ## Post-process section names to extract versions and dates.
+    reDt <- "[[:digit:]]{4}-[[:digit:]]{2}-[[:digit:]]{2}"
+    rEnd <- "[[:punct:][:space:]]*$"
+    re_d1 <- sprintf(paste0("^.*(%s)", rEnd), reDt)
+    ## or ending with '(YYYY-MM-DD, <note>)'
+    re_d2 <- sprintf(paste0("^.*\\((%s)[[:punct:]] .*\\)", rEnd), reDt)
+    nms <- db[, 1L]
+    ind <- grepl(re_v, nms, ignore.case = TRUE)
+    if(!all(ind))
+        warning("Cannot extract version info from the following section titles:\n",
+                paste(unique(nms[!ind]), collapse = "  "))
+
+    .make_news_db(cbind(ifelse(ind,
+                               sub(re_v, "\\2", nms, ignore.case = TRUE),
+                               NA_character_),
+                        ifelse(grepl(re_d1, nms, perl = TRUE),
+                               sub(re_d1, "\\1", nms, perl = TRUE),
+                               ifelse(grepl(re_d2, nms, perl = TRUE),
+                                      sub(re_d2, "\\1", nms, perl = TRUE),
+                                      NA_character_)),
+                        db[, 2L],
+                        Text = sub("\n*$", "", db[, 3L]),
+                        HTML = db[, 4L]),
+                  NULL,
+                  "news_db_from_md")
+}
+
+format.news_db_from_md <-
+function(x, ...)
+{
+    do_vchunk <- function(vchunk) {
+        z <- unlist(Map(c, vchunk$Category, vchunk$Text),
+                    use.names = FALSE)
+        z[nzchar(z)]
+    }
+
+    vchunks <- split(x, x$Version)
+    ## Re-order according to decreasing version.
+    vchunks <- vchunks[order(numeric_version(names(vchunks),
+                                             strict = FALSE),
+                             decreasing = TRUE)]
+    if(!length(vchunks))
+        return(character())
+
+    dates <- sapply(vchunks, function(v) v$Date[1L])
+    vheaders <-
+        format(sprintf("Changes in version %s%s",
+                       names(vchunks),
+                       ifelse(is.na(dates), "",
+                              sprintf(" (%s)", dates))),
+               justify = "centre", width = 72L)
+
+    Map(c, vheaders, lapply(vchunks, do_vchunk),
+        USE.NAMES = FALSE)
+}
+
+.news_db_has_no_bad_entries <-
+function(x)
+{
+    (is.null(bad <- attr(x, "bad")) ||
+     (length(bad) == NROW(x)) && !any(bad))
+}
